
TIMESTAMP = 'timestamp'
PREVIOUS_STATE = 'previous_state'
NEW_STATE = 'new_state'

STATE_HISTORY_KEYS = (TIMESTAMP, PREVIOUS_STATE, NEW_STATE)

INITIAL_STATE = 'initial_state'
TRANSITIONS = 'transitions'
STATE_MACHINE = 'state_machine'

CONFIG_MANDATORY_KEYS = (INITIAL_STATE, TRANSITIONS, STATE_MACHINE)

DEFAULT = 'default'